@@ -1,5 +1,5 @@
 /******************************************************************************
- * Copyright (c) 2016-2017, Bradley J Chambers (brad.chambers@gmail.com)
+ * Copyright (c) 2016-2017, 2020 Bradley J Chambers (brad.chambers@gmail.com)
  *
  * All rights reserved.
  *
@@ -63,23 +63,19 @@ namespace pdal
 using namespace Dimension;
 using namespace Eigen;
 
-static StaticPluginInfo const s_info
-{
-    "filters.smrf",
-    "Simple Morphological Filter (Pingel et al., 2013)",
-    "http://pdal.io/stages/filters.smrf.html"
-};
+static StaticPluginInfo const s_info{
+    "filters.smrf", "Simple Morphological Filter (Pingel et al., 2013)",
+    "http://pdal.io/stages/filters.smrf.html"};
 
 // Without the cast, MSVC complains, which is ridiculous when the output
 // is, by definition, an int.
 namespace
 {
-template<typename T>
-T ceil(double d)
+template <typename T> T ceil(double d)
 {
     return static_cast<T>(std::ceil(d));
 }
-}
+} // namespace
 
 CREATE_STATIC_STAGE(SMRFilter, s_info)
 
@@ -94,13 +90,12 @@ struct SMRArgs
     std::string m_dir;
     std::vector<DimRange> m_ignored;
     StringList m_returns;
+    StringList m_classbits;
 };
 
-SMRFilter::SMRFilter() : m_args(new SMRArgs)
-{}
+SMRFilter::SMRFilter() : m_args(new SMRArgs) {}
 
-SMRFilter::~SMRFilter()
-{}
+SMRFilter::~SMRFilter() {}
 
 std::string SMRFilter::getName() const
 {
@@ -119,6 +114,8 @@ void SMRFilter::addArgs(ProgramArgs& args)
     args.add("ignore", "Ignore values", m_args->m_ignored);
     args.add("returns", "Include last returns?", m_args->m_returns,
              {"last", "only"});
+    args.add("classbits", "Ignore synthetic|keypoint|withheld classification bits?",
+             m_args->m_classbits, {""});
 }
 
 void SMRFilter::addDimensions(PointLayoutPtr layout)
@@ -133,7 +130,7 @@ void SMRFilter::prepared(PointTableRef table)
     for (auto& r : m_args->m_ignored)
     {
         r.m_id = layout->findDim(r.m_name);
-        if (r.m_id == Dimension::Id::Unknown)
+        if (r.m_id == Id::Unknown)
             throwError("Invalid dimension name in 'ignored' option: '" +
                        r.m_name + "'.");
     }
@@ -149,8 +146,8 @@ void SMRFilter::prepared(PointTableRef table)
             }
         }
 
-        if (!layout->hasDim(Dimension::Id::ReturnNumber) ||
-            !layout->hasDim(Dimension::Id::NumberOfReturns))
+        if (!layout->hasDim(Id::ReturnNumber) ||
+            !layout->hasDim(Id::NumberOfReturns))
         {
             log()->get(LogLevel::Warning) << "Could not find ReturnNumber and "
                                              "NumberOfReturns. Skipping "
@@ -185,17 +182,23 @@ PointViewSet SMRFilter::run(PointViewPtr view)
         Segmentation::ignoreDimRanges(m_args->m_ignored, view, keptView,
                                       ignoredView);
 
+    PointViewPtr syntheticView = keptView->makeNew();
+    PointViewPtr realView = keptView->makeNew();
+    if (!m_args->m_classbits.size())
+        realView->append(*keptView);
+    else
+        Segmentation::ignoreClassBits(keptView, realView, syntheticView,
+                                      m_args->m_classbits);
+
     // Check for 0's in ReturnNumber and NumberOfReturns
     bool nrOneZero(false);
     bool rnOneZero(false);
     bool nrAllZero(true);
     bool rnAllZero(true);
-    for (PointId i = 0; i < keptView->size(); ++i)
+    for (PointId i = 0; i < realView->size(); ++i)
     {
-        uint8_t nr =
-            keptView->getFieldAs<uint8_t>(Dimension::Id::NumberOfReturns, i);
-        uint8_t rn =
-            keptView->getFieldAs<uint8_t>(Dimension::Id::ReturnNumber, i);
+        uint8_t nr = realView->getFieldAs<uint8_t>(Id::NumberOfReturns, i);
+        uint8_t rn = realView->getFieldAs<uint8_t>(Id::ReturnNumber, i);
         if ((nr == 0) && !nrOneZero)
             nrOneZero = true;
         if ((rn == 0) && !rnOneZero)
@@ -212,39 +215,42 @@ PointViewSet SMRFilter::run(PointViewPtr view)
                    "1.");
 
     // Segment kept view into two views
-    PointViewPtr firstView = keptView->makeNew();
-    PointViewPtr secondView = keptView->makeNew();
+    PointViewPtr inlierView = realView->makeNew();
+    PointViewPtr outlierView = realView->makeNew();
     if (nrAllZero && rnAllZero)
     {
         log()->get(LogLevel::Warning)
             << "Both NumberOfReturns and ReturnNumber are filled with 0's. "
                "Proceeding without any further return filtering.\n";
-        firstView->append(*keptView);
+        inlierView->append(*realView);
     }
     else
     {
-        Segmentation::segmentReturns(keptView, firstView, secondView,
+        Segmentation::segmentReturns(realView, inlierView, outlierView,
                                      m_args->m_returns);
+        ignoredView->append(*outlierView);
     }
 
-    if (!firstView->size())
+    if (!inlierView->size())
     {
         throwError("No returns to process.");
     }
 
-    for (PointId i = 0; i < secondView->size(); ++i)
-        secondView->setField(Dimension::Id::Classification, i, 1);
+    // Classify remaining points with value of 1. SMRF processing will mark
+    // ground returns as 2.
+    for (PointId i = 0; i < inlierView->size(); ++i)
+        inlierView->setField(Id::Classification, i, ClassLabel::Unclassified);
 
-    m_srs = firstView->spatialReference();
+    m_srs = inlierView->spatialReference();
 
-    calculateBounds(*firstView, m_bounds);
+    calculateBounds(*inlierView, m_bounds);
     m_cols = static_cast<int>(
         ((m_bounds.maxx - m_bounds.minx) / m_args->m_cell) + 1);
     m_rows = static_cast<int>(
         ((m_bounds.maxy - m_bounds.miny) / m_args->m_cell) + 1);
 
     // Create raster of minimum Z values per element.
-    std::vector<double> ZImin = createZImin(firstView);
+    std::vector<double> ZImin = createZImin(inlierView);
 
     // Create raster mask of pixels containing low outlier points.
     std::vector<int> Low = createLowMask(ZImin);
@@ -264,16 +270,18 @@ PointViewSet SMRFilter::run(PointViewPtr view)
     // original ZImin (not ZInet), however the net cut mask will still force
     // interpolation at these pixels.
     std::vector<double> ZIpro =
-        createZIpro(firstView, ZImin, Low, isNetCell, Obj);
+        createZIpro(inlierView, ZImin, Low, isNetCell, Obj);
 
     // Classify ground returns by comparing elevation values to the provisional
     // DEM.
-    classifyGround(firstView, ZIpro);
+    classifyGround(inlierView, ZIpro);
 
     PointViewPtr outView = view->makeNew();
+    // ignoredView is appended to the output untouched.
     outView->append(*ignoredView);
-    outView->append(*secondView);
-    outView->append(*firstView);
+    // inlierView is appended to the output, the only PointView whose
+    // classifications may have been altered.
+    outView->append(*inlierView);
     viewSet.insert(outView);
 
     return viewSet;
@@ -362,9 +370,9 @@ void SMRFilter::classifyGround(PointViewPtr view, std::vector<double>& ZIpro)
         // vertical distance between each LIDAR point and the provisional
         // DEM, and applying a threshold calculation."
         if (std::fabs(ZIpro[c * m_rows + r] - z) > thresh(r, c))
-            view->setField(Id::Classification, i, 1);
+            view->setField(Id::Classification, i, ClassLabel::Unclassified);
         else
-            view->setField(Id::Classification, i, 2);
+            view->setField(Id::Classification, i, ClassLabel::Ground);
     }
 }
 
@@ -448,8 +456,6 @@ std::vector<int> SMRFilter::createObjMask(std::vector<double> const& ZImin)
 
 std::vector<double> SMRFilter::createZImin(PointViewPtr view)
 {
-    using namespace Dimension;
-
     // "As with many other ground filtering algorithms, the first step is
     // generation of ZImin from the cell size parameter and the extent of the
     // data."
@@ -609,7 +615,7 @@ std::vector<double> SMRFilter::knnfill(PointViewPtr view,
             double x = m_bounds.minx + (c + 0.5) * m_args->m_cell;
             double y = m_bounds.miny + (r + 0.5) * m_args->m_cell;
             int k = 8;
-            std::vector<PointId> neighbors(k);
+            PointIdList neighbors(k);
             std::vector<double> sqr_dists(k);
             kdi.knnSearch(x, y, k, &neighbors, &sqr_dists);
 
