@@ -116,6 +116,66 @@ TEST(CropFilterTest, test_crop)
     EXPECT_NEAR(avgZ, 500.00000, delZ);
 }
 
+TEST(CropFilterTest, test_crop_3d)
+{
+    BOX3D srcBounds(0.0, 0.0, 0.0, 10.0, 100.0, 1000.0);
+    Options opts;
+    opts.add("bounds", srcBounds);
+    opts.add("count", 1000);
+    opts.add("mode", "ramp");
+    FauxReader reader;
+    reader.setOptions(opts);
+
+    // crop the window to 20% the size in each dimension
+    BOX3D dstBounds(2.0, 20.0, 200.0, 4.0, 40.0, 400.0);
+    Options cropOpts;
+    cropOpts.add("bounds", dstBounds);
+
+    CropFilter filter;
+    filter.setOptions(cropOpts);
+    filter.setInput(reader);
+
+    Options statOpts;
+
+    StatsFilter stats;
+    stats.setOptions(statOpts);
+    stats.setInput(filter);
+
+    PointTable table;
+    stats.prepare(table);
+    PointViewSet viewSet = stats.execute(table);
+    EXPECT_EQ(viewSet.size(), 1u);
+    PointViewPtr buf = *viewSet.begin();
+
+    const stats::Summary& statsX = stats.getStats(Dimension::Id::X);
+    const stats::Summary& statsY = stats.getStats(Dimension::Id::Y);
+    const stats::Summary& statsZ = stats.getStats(Dimension::Id::Z);
+    EXPECT_EQ(buf->size(), 200u);
+
+    const double minX = statsX.minimum();
+    const double minY = statsY.minimum();
+    const double minZ = statsZ.minimum();
+    const double maxX = statsX.maximum();
+    const double maxY = statsY.maximum();
+    const double maxZ = statsZ.maximum();
+    const double avgX = statsX.average();
+    const double avgY = statsY.average();
+    const double avgZ = statsZ.average();
+
+    const double delX = 10.0 / 999.0 * 100.0;
+    const double delY = 100.0 / 999.0 * 100.0;
+    const double delZ = 1000.0 / 999.0 * 100.0;
+
+    EXPECT_NEAR(minX, 2.0, delX);
+    EXPECT_NEAR(minY, 20.0, delY);
+    EXPECT_NEAR(minZ, 200.0, delZ);
+    EXPECT_NEAR(maxX, 4.0, delX);
+    EXPECT_NEAR(maxY, 40.0, delY);
+    EXPECT_NEAR(maxZ, 400.0, delZ);
+    EXPECT_NEAR(avgX, 3.00000, delX);
+    EXPECT_NEAR(avgY, 30.00000, delY);
+    EXPECT_NEAR(avgZ, 300.00000, delZ);
+}
 
 TEST(CropFilterTest, test_crop_polygon)
 {
